/*
 * FILE:
 * action.c
 *
 * FUNCTION:
 * Handler for a button click.
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */

#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>

#include "action.h"
#include "builder.h"
#include "dui-initdb.h"
#include "window.h"
#include "perr.h"
#include "report.h"
#include "signal.h"
#include "util.h"

typedef struct DuiRefresh_s DuiRefresh;

struct DuiAction_s
{
	DuiWindow *window;  /* the window in which we live */

	/* the query builder */
	SqlBuilder *sql_builder;

	/* a portion of the sql query */
	SqlBuilderQType qtype;
	GList *update_terms;
	GList *where_terms;

	/* cache of most recent sql query string; used for rerunning the query. */
	char * query_string;

	/* database connection on which to send the query */
	DuiDBConnection *db_conn;
	
	/* list of widgets whose values we pput into kvp */
	GList *keys;
	
	/* the report that will report the results of the query */
	char * report_name;
	DuiReport * report;

	/* list of gtk signals that can trigger this action. */
	GList *signals;

	/* list of windows that need to be refreshed by this action */
	GList *refresh_list;

};

struct DuiRefresh_s
{
	char * windowname;
	DuiReport *window;
};


static void dui_action_signal_handler (GtkObject *obj, DuiAction *act);

/* =================================================================== */
/* the following struct defines how widget fields (read as input) are
 * mapped to database fields (used to formulate the query)
 */

typedef struct DuiQueryFieldMap_s DuiQueryFieldMap;

struct DuiQueryFieldMap_s 
{
	DuiAction *act;
	
	/* the next four members specify the widget from which
	 * we will be reading data. */
	char * widgetname;
	GtkWidget * widget; 
	int column;
	int row;
	char * key;   /* the key side of a kvp pair */
	char * value; /* a hard-coded value to use, if the widget 
                  * is not present */

	/* the next three members specify how that widget's
	 * data is mapped to a database field. */
	char * fieldtype;
	char * fieldname;
	char * compareop;
};


/* =================================================================== */

static void
terms_destroy (GList *list)
{
	GList *node;
	
	/* remove the insert/update terms */
	for (node=list; node; node=node->next)
	{
		DuiQueryFieldMap *qfm = node->data;

		g_free (qfm->widgetname);
		g_free (qfm->fieldname);
		g_free (qfm->fieldtype);
		g_free (qfm->compareop);
		g_free (qfm->key);
		g_free (qfm->value);
		g_free (qfm);
	}
	g_list_free (list);
}

/* =================================================================== */

DuiAction * 
dui_action_new (DuiWindow *win)
{
	DuiAction *act;

	if (!win) return NULL;

	act = g_new (DuiAction, 1);
	act->window = win;

	act->signals = NULL;

	act->sql_builder = NULL;

	act->qtype = 0;
	act->update_terms = NULL;
	act->where_terms = NULL;
	act->query_string = NULL;
	act->db_conn = NULL;

	act->keys = NULL;

	act->report_name = NULL;
	act->report = NULL;
	act->refresh_list = NULL;

	dui_window_add_action (win, act);
	return act;
}

void 
dui_action_destroy (DuiAction * act)
{
	GList *node;
	if (!act) return;

	act->window = NULL;

        sql_builder_destroy(act->sql_builder);
        act->sql_builder = NULL;
	act->qtype = 0;

	/* remove the insert/update terms */
	terms_destroy (act->update_terms);
	terms_destroy (act->where_terms);
	terms_destroy (act->keys);
	act->update_terms = NULL;
	act->where_terms = NULL;
	act->keys = NULL;

	g_free(act->report_name);
	act->report_name = NULL;
	act->report = NULL;

	/* the signals themselves will be reaped by window */
	g_list_free (act->signals);
	act->signals = NULL;

	act->db_conn = NULL;

	for (node=act->refresh_list; node; node=node->next)
	{
		DuiRefresh *rsh = node->data;
		g_free (rsh->windowname);
		g_free (rsh);
	}
	g_list_free (act->refresh_list);
	g_free (act);
}

/* =================================================================== */

void 
dui_action_set_report (DuiAction * act, const char * dpyname)
{
	if (!act) return;
	if (act->report_name) g_free (act->report_name);
	act->report_name = g_strdup (dpyname);
}
	
/* =================================================================== */


void 
dui_action_add_refresh (DuiAction *act, const char * name)
{
	DuiRefresh *rsh;
	if (!act) return;
	
	rsh = g_new (DuiRefresh, 1);
	rsh->windowname = g_strdup (name);
	rsh->window = NULL;

	act->refresh_list = g_list_prepend (act->refresh_list, rsh);
}

/* =================================================================== */

void 
dui_action_set_db_connection (DuiAction * act, DuiDBConnection *dbc)
{
	if (!act) return;
	act->db_conn = dbc;
}
	
DuiDBConnection *
dui_action_get_db_connection (DuiAction * act)
{
	if (!act) return NULL;
	return act->db_conn;
}

/* =================================================================== */

void 
dui_action_add_signal (DuiAction *act, const char *objectname, const char *signalname)
{
	DuiSignal *sig;

	if (!act || !objectname || !signalname) return;
	
	sig = dui_signal_new (act->window, objectname, signalname, "submit_form");
	act->signals = g_list_append (act->signals, sig);
}
			
/* =================================================================== */

void 
dui_action_set_table (DuiAction *act, const char *tables, 
                                      const char * sql_querytype)
{

	if (!act || !tables || !sql_querytype) return;

	/* set up the sql statement builder so that the next set of
	 * of xml stanzas set up a valid sql statement
	 */
	if (!strcasecmp (sql_querytype, "select")) act->qtype = SQL_SELECT;
	else if (!strcasecmp (sql_querytype, "select")) act->qtype = SQL_SELECT;
	else if (!strcasecmp (sql_querytype, "insert")) act->qtype = SQL_INSERT;
	else if (!strcasecmp (sql_querytype, "update")) act->qtype = SQL_UPDATE;
	else if (!strcasecmp (sql_querytype, "delete")) act->qtype = SQL_DELETE;

	act->sql_builder = sql_builder_new ();
	sql_builder_table (act->sql_builder, tables, act->qtype);

}

/* =================================================================== */

void 
dui_action_add_select_term (DuiAction * act, 
                           const char *field_name)
{
	if (!act || !field_name) return;
	sql_builder_set_str (act->sql_builder, field_name, NULL);
}


/* =================================================================== */

void 
dui_action_add_update_term (DuiAction * act, 
                           const char * widname,
                           int colnum,
                           const char *field_name,
                           const char *key,
                           const char *value)
{
	DuiQueryFieldMap *qfm;

	if (!act || !field_name) return;

	/* save them up for when we need them */
	qfm = g_new (DuiQueryFieldMap, 1);
	qfm->act = act;
	qfm->widgetname = g_strdup (widname),
	qfm->widget = NULL,
	qfm->column = colnum;
	qfm->row = -1;
	qfm->fieldname = g_strdup (field_name);
	qfm->fieldtype = NULL;
	qfm->compareop = NULL;
	qfm->key = g_strdup (key);
	qfm->value = g_strdup (value);
	
	act->update_terms = g_list_append (act->update_terms, qfm);
}

/* =================================================================== */

void 
dui_action_add_where_term (DuiAction * act, 
                           const char * widname,
                           int colnum,
                           const char *field_name,
                           const char *field_type,
                           const char *compare_op,
                           const char *key,
									const char *value)
{
	DuiQueryFieldMap *qfm;

	if (!act || !field_name) return;

	/* save them up for when we need them */
	qfm = g_new (DuiQueryFieldMap, 1);
	qfm->act = act;
	qfm->widgetname = g_strdup (widname);
	qfm->widget = NULL;
	qfm->column = colnum;
	qfm->row = -1;
	qfm->fieldname = g_strdup (field_name);
	qfm->fieldtype = g_strdup (field_type);
	qfm->compareop = g_strdup (compare_op);
	qfm->key = g_strdup (key);
	qfm->value = g_strdup (value);

	act->where_terms = g_list_append (act->where_terms, qfm);
}

/* =================================================================== */

void 
dui_action_add_kvp_term (DuiAction * act, 
                           const char * widname,
                           int colnum,
                           const char *key,
									const char *value)
{
	DuiQueryFieldMap *qfm;

	if (!act) return;

	/* save them up for when we need them */
	qfm = g_new (DuiQueryFieldMap, 1);
	qfm->act = act;
	qfm->widgetname = g_strdup (widname);
	qfm->widget = NULL;
	qfm->column = colnum;
	qfm->row = -1;
	qfm->fieldname = NULL;
	qfm->fieldtype = NULL;
	qfm->compareop = NULL;
	qfm->key = g_strdup (key);
	qfm->value = g_strdup (value);

	act->keys = g_list_append (act->keys, qfm);
}

/* =================================================================== */

static const char *
get_value (GtkWidget *widget, int row, int col)
{
	const char * val = NULL;

	if (!widget) return NULL;   /* this shouldn't happen, really */

	/* supported widgets, in alpha order */
	if (GTK_IS_CLIST(widget))
	{
		if (0 > row) return NULL;
		gtk_clist_get_text (GTK_CLIST(widget), row, col, (gchar **) &val);
	}
	else
	if (GTK_IS_ENTRY(widget))
	{
	  	val = gtk_entry_get_text (GTK_ENTRY(widget));
	}
	else
	if (GTK_IS_LABEL(widget))
	{
	  	gtk_label_get (GTK_LABEL(widget), (gchar **) &val);
	}
	else
	if (GTK_IS_TEXT(widget))
	{
	  	val = xxxgtk_text_get_text (GTK_TEXT(widget));
	}
	else
	{
	  	PERR ("unsupported <where> or <update> widget type %s",
			gtk_type_name(GTK_OBJECT_TYPE(widget)));
	}
	return val;
}

/* =================================================================== */

static void
final_setup (DuiAction *act)
{
	GList *node;
	DuiInterface *dui;

	/* locate the reports the action will use, and plug them in. */
	dui = dui_window_get_interface (act->window);
	act->report = dui_interface_find_report_by_name (dui, act->report_name);

	if (act->report_name && !act->report) 
	{
		PERR ("can't find the report \'%s\'", act->report_name);
	}

	/* do the same thing for the refresh list */
	for (node=act->refresh_list; node; node=node->next)
	{
		DuiRefresh *rsh = node->data;
		rsh->window = dui_interface_find_report_by_name (dui, rsh->windowname);

		if (rsh->windowname && !rsh->window) 
		{
			PERR ("can't find the window \'%s\'", rsh->windowname);
		}
	}
}

/* =================================================================== */

static void
get_clist_row (GtkCList *clist,
               gint            row,
               gint            column,
               GdkEvent       *event,
               DuiQueryFieldMap *qfm)
{
	qfm->row = row;
}

static void
add_watchers (DuiAction *act, GList *terms)
{
	GList *node;

	for (node=terms; node; node=node->next)
	{
		DuiQueryFieldMap *qfm = node->data;
		if (qfm->widgetname)
		{
			qfm->widget =  dui_window_get_widget (act->window, qfm->widgetname);

			/* There's no way to ask the clist what row has been selected,
			 * so we have to listen for events instead. */
			if ((GTK_IS_CLIST (qfm->widget)) ||
			    (GTK_IS_CTREE(qfm->widget)))
			{
				gtk_signal_connect(GTK_OBJECT(qfm->widget), "select_row",
		   	           GTK_SIGNAL_FUNC(get_clist_row), qfm);
			}
		}
	}
}

void
dui_action_do_realize (DuiAction *act)
{
	GList *node;
	if (!act) return;

	ENTER ("(act=%p)", act);
	/* Don't bother if the parent window doesn't exist. */
	/* Technically, this is an error, it shouldn't happen ... */
	if (0 == dui_window_is_realized(act->window)) return;

	if (act->report_name && !act->report) final_setup (act);

	for (node=act->signals; node; node=node->next)
	{
		GtkWidget *w;
		DuiSignal *sig = node->data;
		w = dui_window_get_widget (act->window, dui_signal_get_widget(sig));
		/* might not be a widget if the root window doesn't exist */
		if (!w) continue;  

		dui_signal_connect (sig, GTK_OBJECT(w),
		         GTK_SIGNAL_FUNC(dui_action_signal_handler), act);
	}

	add_watchers (act, act->where_terms);
	add_watchers (act, act->update_terms);
	add_watchers (act, act->keys);

	LEAVE ("(act=%p)", act);
}

/* =================================================================== */

int
dui_action_is_realized (DuiAction *act)
{
	if (!act) return 0;

	return dui_window_is_realized(act->window);
}

/* =================================================================== */
/* return NULL if the field is whitespace (blank, tab, formfeed etc.)  */

static const char *
whitespace_filter (const char * val)
{
	size_t len;
	if (!val) return NULL;

	len = strspn (val, "\a\b\t\n\v\f\r ");
	if (0 == val[len]) return NULL;
	return val+len;
}


/* =================================================================== */

static const char *
qfm_get_value (DuiQueryFieldMap *qfm)
{
	const char * fieldval;
	if (qfm->widget)
	{
		fieldval = get_value (qfm->widget, qfm->row, qfm->column);
	}
	else
	if (qfm->key)
	{
		DuiInterface *dui = dui_window_get_interface (qfm->act->window);
		fieldval = dui_interface_kvp_lookup (dui, qfm->key);
	}
	else
	{
		fieldval = qfm->value;
	}
	return fieldval;
}

static void
dui_action_signal_handler (GtkObject *obj, DuiAction *act)
{
	const char * stmt;
	DuiDBRecordSet *recs = NULL;
	GList *node;
	
	ENTER ("(act=%p)", act);

	/* loop over keys, fetch values and stuff them into kvp */
	for (node = act->keys; node; node=node->next)
	{
		DuiInterface *dui;
		DuiQueryFieldMap *qfm = node->data;
		const char * fieldval;
		if (qfm->widget)
		{
			fieldval = get_value (qfm->widget, qfm->row, qfm->column);
		}
		else
		{
			fieldval = qfm->value;
		}
		
		dui = dui_window_get_interface (act->window);
		dui_interface_kvp_insert (dui, qfm->key, fieldval);
	}
	
	/* If we have data we need to query, then do the query.
	 * (This is the usual case.  Sometimes we have no data, e.g.
	 * when setting a form to a constant value e.g. to initialize it)
	 */
	if (act->sql_builder)
	{
		GList *node;
		SqlBuilder *dupe;
		short have_where_terms = 0;

		g_free (act->query_string);
		act->query_string = NULL;

		/* make a copy, add the appropriate where clauses */
		dupe = sql_builder_copy (act->sql_builder);
	
		for (node = act->update_terms; node; node=node->next)
		{
			DuiQueryFieldMap *qfm = node->data;
			const char * fieldval = qfm_get_value (qfm);
			
			sql_builder_set_str (dupe, qfm->fieldname, fieldval);
		}
	
		for (node = act->where_terms; node; node=node->next)
		{
			DuiQueryFieldMap *qfm = node->data;
			const char * fieldval = qfm_get_value (qfm);
	
			/* fields left blank don't match */
			fieldval = whitespace_filter (fieldval);
			if (fieldval)
			{
				sql_builder_where_str (dupe, qfm->fieldname, 
					fieldval, qfm->compareop);
				have_where_terms = 1;
			}
		}
	
		/* Perform the query only if its not a wild-card
		 * update or delete.  Wildcard updates/deletes are baaad. */
		if (have_where_terms || ((SQL_UPDATE != act->qtype) && 
		                         (SQL_DELETE != act->qtype)))
		{
			/* get the actual sql statement */
			stmt = sql_builder_query (dupe);
			act->query_string = g_strdup (stmt);
	
			/* now run the query */
			if (act->db_conn)
			{
				recs = dui_connection_exec(act->db_conn, stmt);
			}
		}	
		sql_builder_destroy (dupe);
	}

	/* now report the results */
	dui_report_set_last_action (act->report, act);
	dui_report_show_data (act->report, recs);

	/* don't need it no more */
	dui_recordset_free (recs);

	/* refresh all windows that show data modified by this query */
	for (node=act->refresh_list; node; node=node->next)
	{
		DuiRefresh *rsh = node->data;
		dui_report_refresh (rsh->window);
	}
	LEAVE ("(act=%p)", act);
}

/* =================================================================== */

void
dui_action_rerun_last_query (DuiAction *act)
{
	DuiDBRecordSet *recs = NULL;

	if (!act) return;

	if (!act->db_conn || !act->query_string) return;

	ENTER ("(act=%p)", act);

	recs = dui_connection_exec(act->db_conn, act->query_string);

	/* display the results of the query */
	dui_report_show_data (act->report, recs);

	dui_recordset_free (recs);
	LEAVE ("(act=%p)", act);
}

/* ========================== END OF FILE ============================ */
