/*
 * FILE:
 * readfile.c
 *
 * FUNCTION:
 * parse an XML file that specifies the glade-to-database coupling
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */

#include <string.h>
#include <stdlib.h>

#include <glib.h>

#include <parser.h>
#include <gnome.h>

#include "perr.h"
#include "dui-initdb.h"

#include "action.h"
#include "readfile.h"
#include "report.h"
#include "signal.h"
#include "window.h"

struct DuiInterface_s 
{
	/* application name */
	char * app_name;

	/* XML file parser */
	xmlSAXHandler sax;
	char * xml_filename;

	/* database info */
	char * provider;
	char * dbname;
	char * hostname;
	char * username;

	/* current database connection that is being used */
	DuiDBConnection *db_conn;
	
	/* current glade window that is being processed */
	DuiWindow *window;

	/* current action that is being processed */
	DuiAction  *action;

	/* current report that is being processed */
	DuiReport *report;
	
	/* collection of handlers */
	GList *root_windows;
	GList *action_handlers;
	GList *report_handlers;

	/* set of key-value pairs */ 
	GHashTable *kvp; 
};


#define SYNTAX(x,args...) USERERR ("Syntax Error: " x, ##args)

/* =================================================================== */
/* a really cheap and dirty thing .. */

static const char * 
de_escape (const char *val)
{
	if (!val) return NULL;

	if (!strcmp (val, "&gt;")) return ">";
	if (!strcmp (val, "&lt;")) return "<";

	return val;
}

/* =================================================================== */

#define CASE_STRDUP(STR,MEMB)                          \
	if (!strcmp (attrs[i], STR)) {                 \
		MEMB = g_strdup(attrs[i+1]);           \
	} else

#define CASE(STR,MEMB)                                 \
	if (!strcmp (attrs[i], STR)) {                 \
		MEMB = attrs[i+1];                     \
	} else

#define CASE_UNKNOWN                                 \
	{                                                 \
		SYNTAX ("unknown %s attribute %s", eltname, attrs[i]); \
	}   

#define GENERIC_CHECKS                               \
	if (!dui->window)                                 \
	{                                                 \
		SYNTAX ("must specify a glade file first!");   \
		return;                                        \
	}                                                 \
                                                     \
	if (!attrs)                                       \
	{                                                 \
		SYNTAX ("%s must have attributes!", eltname);  \
		return;                                        \
	}

#define CHECK_NEST(fld,prnt)                         \
	if (!dui->fld)                                    \
	{                                                 \
		SYNTAX ("%s can only occur within a " prnt "!", eltname);\
		return;                                        \
	}

#define REQUIRE(var,what)                            \
	if (NULL == var)                                  \
	{                                                 \
		SYNTAX ("%s must specify " what, eltname);     \
		return;                                        \
	}

#define REQUIRE2(vara,varb,what)                     \
	if ((NULL == vara) && (NULL == varb))             \
	{                                                 \
		SYNTAX ("%s must specify " what, eltname);     \
		return;                                        \
	}

#define REQUIRE3(vara,varb,varc,what)                \
	if ((NULL == vara) && (NULL == varb) && (NULL == varc))   \
	{                                                 \
		SYNTAX ("%s must specify " what, eltname);     \
		return;                                        \
	}

#define REQUIRE4(vara,varb,varc,vard,what)           \
	if ((NULL==vara) && (NULL==varb) && (NULL==varc) && (NULL==vard))   \
	{                                                 \
		SYNTAX ("%s must specify " what, eltname);     \
		return;                                        \
	}


/* =================================================================== */

static void 
handle_report (DuiInterface *dui, const xmlChar **attrs)
{
	DuiReport *rpt;
	static const char * eltname = "<report>";
	int i;

	if (!dui->window) 
	{
		SYNTAX ("must specify a glade file first!");
		return;
	}

	if (attrs)
	{
		i = 0;
		while (attrs[i])
		{
			if (0) {}
			CASE_UNKNOWN ;
			i += 2;
		}
	}

	/* ------------------------------------------------------------------ */
	/* Set up the widget so that when its clicked or whatever, 
	 * we run our sql statement.
	 */
	rpt = dui_report_new (dui->window);
	dui->report = rpt;
	dui->report_handlers = g_list_append (dui->report_handlers, rpt);
}

static void 
handle_report_end (DuiInterface *dui)
{
	if (!dui->report) return;

	dui->report = NULL;
}

/* =================================================================== */

static void 
handle_row (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<row>";
	int i;
	const char * name = NULL;
	const char * widgetname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (report, "<report>");

	i = 0;
	while (attrs[i])
	{
		CASE ("name",   name)
		CASE ("widget", widgetname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (widgetname, "a widget!");
	dui_report_set_row_widget (dui->report, widgetname);

}

/* =================================================================== */

static void 
handle_entry (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<entry>";
	int i;
	const char * name = NULL;
	const char * widgetname = NULL;
	const char * fieldname = NULL;
	const char * column = NULL;
	const char * value = NULL;
	const char * key = NULL;
	int col = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (report, "<report>");

	i = 0;
	while (attrs[i])
	{
		CASE ("name",   name)
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("field", fieldname)
		CASE ("value", value)
		CASE ("key", value)
		CASE_UNKNOWN ;
		i += 2;
	}

	// hack alert -- fixme -- 
	// a widget is not required if we are nested inside a row;
	// otherwise a widget is mandatory
	// REQUIRE (widgetname, "a widget!");
	// 
	REQUIRE3 (fieldname, value, key, "a field or a key or a value!");

	if (column) col = atoi (column);

	dui_report_add_column (dui->report, widgetname, col, fieldname, key, value);
}


/* =================================================================== */
/* ********************* form handlers below ************************* */
/* =================================================================== */

/* =================================================================== */

static void 
handle_form (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<form>";
	int i;
	const char * reportname = NULL;
	const char * sql_querytype = NULL;
	const char * sql_tables = NULL;
	DuiAction *act;

	if (attrs)
	{
		i = 0;
		while (attrs[i])
		{
			CASE ("type",   sql_querytype)
			CASE ("tables", sql_tables)
			CASE ("report", reportname)
			CASE_UNKNOWN ;
			i += 2;
		}
	}

	/* ------------------------------------------------------------------ */
	/* Set up the widget so that when its clicked or whatever, 
	 * we run our sql statement.
	 */
	act = dui_action_new (dui->window);
	dui->action = act;
	dui->action_handlers = g_list_append (dui->action_handlers, act);

	dui_action_set_table (act, sql_tables, sql_querytype);
	dui_action_set_report (act, reportname);
	dui_action_set_db_connection (act, dui->db_conn);
}

static void 
handle_form_end (DuiInterface *dui)
{
	if (!dui->action) return;

	dui->action = NULL;
}

/* =================================================================== */

static void 
handle_submit (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<submit>";
	int i;
	const char * widgetname = NULL;
	const char * signalname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("signal", signalname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (widgetname, "a widget!");
	REQUIRE (signalname, "a signal!");

	dui_action_add_signal (dui->action, widgetname, signalname);
}

/* =================================================================== */

static void 
handle_refresh (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<refresh>";
	int i;
	const char * windowname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("window", windowname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (windowname, "a window name!");
	dui_action_add_refresh (dui->action, windowname);
}

/* =================================================================== */

static void 
handle_select (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<select>";
	int i;
	const char * fieldname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("field", fieldname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (fieldname, "a data field!");
	dui_action_add_select_term (dui->action, fieldname);
}

/* =================================================================== */

static void 
handle_insert (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<insert>";
	int i;
	const char * fieldname = NULL;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * key = NULL;
	const char * value = NULL;
	int colnum = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("field",  fieldname)
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("key",    key)
		CASE ("value",  value)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE3 (widgetname, key, value, "a widget or a key or a value!");
	REQUIRE (fieldname, "a data field!");
	
	if (column) colnum = atoi (column);

	/* set up the action so that we fetch values from it */
	dui_action_add_update_term(dui->action, widgetname, 
                   colnum, fieldname, key, value);
}

/* =================================================================== */

static void 
handle_where (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<where>";
	int i;
	const char * widgetname = NULL;
	const char * fieldname = NULL;
	const char * fieldtype = NULL;
	const char * compareop = NULL;
	const char * column = NULL;
	const char * key = NULL;
	const char * value = NULL;
	int colnum = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("field",  fieldname)
		CASE ("op",     compareop)
		CASE ("column", column)
		CASE ("key",    value)
		CASE ("value",  value)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* require either value or a key or a widget ... */
	REQUIRE3 (widgetname, key, value, "a widget or a key or a value!");
	REQUIRE (fieldname, "a data field!");

	if (column) colnum = atoi (column);

	/* set up the action so that we fetch values from it */
	compareop = de_escape (compareop);
	dui_action_add_where_term(dui->action, widgetname, colnum, 
		         fieldname, fieldtype, compareop, key, value);

}

/* =================================================================== */

static void 
handle_wtokey (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<wtokey>";
	int i;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * key = NULL;
	const char * value = NULL;
	int colnum = -1;

	GENERIC_CHECKS ;
	CHECK_NEST (action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("key",    key)
		CASE ("value",  value)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* require either value or widget ... */
	REQUIRE2 (widgetname, value, "a widget or a value!");
	REQUIRE (key, "a key!");

	if (column) colnum = atoi (column);

	/* set up the action so that we fetch values from it */
	dui_action_add_kvp_term(dui->action, widgetname, colnum, key, value);

}

/* =================================================================== */

static void 
handle_sigaction (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<sigaction>";
	int i;
	const char * widgetname = NULL;
	const char * signalname = NULL;
	const char * sigaction = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (window, "<window>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("signal", signalname)
		CASE ("action", sigaction)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (widgetname, "a widget!");
	REQUIRE (signalname, "a valid signal name on the widget!");
	REQUIRE (sigaction,  "an action to perform!");

	dui_signal_new (dui->window, widgetname, signalname, sigaction);
}

/* =================================================================== */

static void 
handle_window (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<window>";
	int i;
	const char * provider = NULL;
	const char * winname = NULL;
	const char * filepath = NULL;
	const char * root_widget = NULL;
	const char * mainattr = NULL;
	int is_main = 0;

	if (!attrs)
	{
		SYNTAX ("<window> must have attributes!");
		return;
	}

	i = 0;
	while (attrs[i])
	{
		CASE ("provider",    provider)
		CASE ("name",        winname)
		CASE ("filename",    filepath)
		CASE ("root_widget", root_widget)
		CASE ("main",        mainattr)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (filepath, "a file path!");
	REQUIRE (root_widget, "a root widget!");

	if (mainattr && !strcasecmp (mainattr, "yes")) is_main = 1;

	dui->window = dui_window_new (winname, dui);
	dui_window_set_glade (dui->window, filepath, root_widget, is_main);

	dui->root_windows = g_list_append (dui->root_windows, dui->window);
}

static void 
handle_window_end (DuiInterface *dui)
{
	dui->window = NULL;
}

/* =================================================================== */

static void 
handle_database (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<database>";
	int i;

	if (!attrs)
	{
		SYNTAX ("<database> must have attributes!");
		return;
	}

	i = 0;
	while (attrs[i])
	{
		CASE_STRDUP ("provider", dui->provider)
		CASE_STRDUP ("dbname",   dui->dbname)
		CASE_STRDUP ("host",     dui->hostname)
		CASE_STRDUP ("username", dui->username)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (dui->provider, "a databasse provider!");
	REQUIRE (dui->dbname,   "a databasse name!");
	REQUIRE (dui->username, "a user name!");

	/* now actually open the database */
	dui->db_conn = dui_connection_new (dui->provider, dui->dbname, dui->username);
	if (NULL == dui->db_conn)
	{
		FATAL ("can't open connection for provider=\"%s\", "
		       "database=\"%s\", username=\"%s\"", 
				 dui->provider, dui->dbname, dui->username);
	}

}

/* =================================================================== */

static void 
handle_dwi (DuiInterface *dui, const xmlChar **attrs)
{
	static const char * eltname = "<dwi>";
	int i;

	i = 0;
	while (attrs[i])
	{
		CASE_STRDUP ("name", dui->app_name)
		CASE_UNKNOWN ;
		i += 2;
	}
}

/* =================================================================== */

static void 
fatal_err (void *user_data, const char *msg, ...)
{
	/* hack laert -- handle the elipses in the warning message */
	// DuiInterface *dui = (DuiInterface *) user_data;
	PERR ("fatal parsing error: %s\n", msg);
}

static void 
err (void *user_data, const char *msg, ...)
{
	// DuiInterface *dui = (DuiInterface *) user_data;
	PERR ("parsing error: %s\n", msg);
}

static void 
warn (void *user_data, const char *msg, ...)
{
	// DuiInterface *dui = (DuiInterface *) user_data;
	PERR ("parse warning: %s\n", msg);
}

static void 
start_doc (void *user_data)
{
	// DuiInterface *dui = (DuiInterface *) user_data;
	PINFO ("document start ptr=%p", user_data);
}

static void 
end_doc (void *user_data)
{
	// DuiInterface *dui = (DuiInterface *) user_data;
	PINFO ("document end");
}

/* =================================================================== */

#define DISPATCH(tok,handler)                  \
	if (!strcmp (name, tok)) {             \
		handler (dui, attrs);          \
	} else

static void 
start_elt (void *user_data, const xmlChar *name, const xmlChar **attrs)
{
	DuiInterface *dui = (DuiInterface *) user_data;


	DISPATCH ("dwi",       handle_dwi)
	DISPATCH ("database",  handle_database)
	DISPATCH ("window",    handle_window)

	DISPATCH ("report",    handle_report)
	DISPATCH ("row",       handle_row)
	DISPATCH ("entry",     handle_entry)
	
	DISPATCH ("form",      handle_form)
	DISPATCH ("select",    handle_select)
	DISPATCH ("insert",    handle_insert)
	DISPATCH ("where",     handle_where)

	DISPATCH ("submit",    handle_submit)
	DISPATCH ("refresh",   handle_refresh)
	DISPATCH ("wtokey",    handle_wtokey)

	DISPATCH ("sigaction", handle_sigaction)
	{
		int i;
		PERR ("unknown element %s\n", name);
		i = 0;
		while (attrs && attrs[i])
		{
			printf ("\tattr=%s\n", attrs[i]);
			i++;
		}
	}
}


static void 
end_elt (void *user_data, const xmlChar *name)
{
	DuiInterface *dui = (DuiInterface *) user_data;
	PINFO ("</%s>", name);

	if (!strcmp (name, "window"))
	{
		handle_window_end (dui);
	}
	else
	if (!strcmp (name, "form"))
	{
		handle_form_end (dui);
	}
	else
	if (!strcmp (name, "report"))
	{
		handle_report_end (dui);
	}
	
}


static void
cdata_cb (void *user_data, const xmlChar *value, int len)
{
#if 0
	DuiInterface *dui = (DuiInterface *) user_data;
	char *p;
	p = g_strndup (value, len);
	printf ("cdata duude len=%d %s\n", len, p);
	g_free (p);
#endif

}

/* =================================================================== */

static guint 
string_hash_func (const char * key)
{
	guint hash = 0;
	int i=0;
	if (!key) return 0;

	while (key[i])
	{
		hash += key[i] << 5 * (i%6);
		i++;
	}
	return hash;
}

static gint 
string_compare_func (const char *a, const char * b)
{
	return (0 == strcmp (a,b));
}


/* =================================================================== */

DuiInterface * 
dui_interface_new (void)
{
	DuiInterface * dui;

	dui = g_new (DuiInterface,1);

	dui->app_name = NULL;

	/* initialize the SAX parser */
	bzero (&dui->sax, sizeof (xmlSAXHandler));
	dui->sax.startDocument = start_doc;
	dui->sax.endDocument = end_doc;
	dui->sax.fatalError = fatal_err;
	dui->sax.error = err;
	dui->sax.warning = warn;
	dui->sax.startElement = start_elt;
	dui->sax.endElement = end_elt;
	dui->sax.characters = cdata_cb;

	dui->xml_filename = NULL;

	/* database stuff */
	dui->provider = NULL;
	dui->dbname = NULL;
	dui->hostname = NULL;
	dui->username = NULL;

	dui->db_conn = NULL;

	/* current state */
	dui->window = NULL;
	dui->action = NULL;
	dui->report = NULL;

	dui->action_handlers = NULL;
	dui->report_handlers = NULL;
	dui->root_windows = NULL;

	dui->kvp = g_hash_table_new ((GHashFunc) string_hash_func, 
	                   (GCompareFunc) string_compare_func);
	return dui;
}


/* =================================================================== */

#define FREE_STR(str)  if(dui->str) g_free (dui->str); dui->str = NULL;

void 
dui_interface_destroy (DuiInterface *dui)
{
	GList *node;

	if (!dui) return;

	FREE_STR (app_name);
	FREE_STR (xml_filename);
	FREE_STR (provider);
	FREE_STR (dbname);
	FREE_STR (hostname);
	FREE_STR (username);

	/* disconnect from the database */
	dui_connection_free (dui->db_conn);
	dui->db_conn = NULL;

	/* current state */
	dui->window = NULL;
	dui->action = NULL;
	dui->report = NULL;

	/* the action and report handlers are deleted when thier 'parent'
	 * window is deleted; we don't need to delete them here. */
	g_list_free (dui->action_handlers);
	g_list_free (dui->report_handlers);

	for (node=dui->root_windows; node; node=node->next)
	{
		DuiWindow *win = node->data;
		dui_window_destroy (win);
	}
	g_list_free (dui->root_windows);

	/* XXX fixme hack alert need to for-each free the keys and values */
	g_hash_table_destroy (dui->kvp);
}

/* =================================================================== */

void
dui_interface_kvp_insert (DuiInterface *dui, const char *key, 
                                             const char *value)
{
	if (!dui || !key || !value) return;
	g_hash_table_insert (dui->kvp, g_strdup(key), g_strdup(value));
}

const char *
dui_interface_kvp_lookup (DuiInterface *dui, const char *key)
{
	if (!dui) return NULL;
	return g_hash_table_lookup (dui->kvp, key);
}

/* =================================================================== */

void 
dui_interface_parse (DuiInterface *dui, const char * filename)
{
	if (!dui || !filename) return;

	if (0 == g_file_exists (filename))
	{
		FATAL ("Can't find DWI interface description file %s\n", filename);
		return;
	}
	dui->xml_filename = g_strdup (filename);
	xmlSAXUserParseFile (&dui->sax, dui, filename);
}

/* =================================================================== */

DuiReport *
dui_interface_find_report_by_name (DuiInterface *dui, const char * name)
{
	GList *node;

	if (!name) return NULL;

	for (node=dui->report_handlers; node; node=node->next)
	{
		DuiReport *rpt = node->data;
		const char *winname = dui_report_get_name (rpt);
		if (winname && !strcmp(winname, name))
		{
			return rpt;
		}
	}
	return NULL;
}

/* =================================================================== */

void 
dui_interface_realize (DuiInterface *dui, int fatal_if_no_main)
{
	GList *node;
	int found_app_window = 0;

	if (!dui) return;

	for (node=dui->root_windows; node; node=node->next)
	{
		DuiWindow *win = node->data;
		int is_app_window = dui_window_is_app_main_window(win);
		if (is_app_window) 
		{
			dui_window_realize (win);
			found_app_window = 1;
		}
	}

	/* Without this test, its just to easy to write a dwi file 
	 * that just hangs there, doing nothing, and you scratch your 
	 * head wondering why ...
	 */
	if (fatal_if_no_main && 0 == found_app_window)
	{
		FATAL ("could not find an application main window!");
	}

}

/* ========================== END OF FILE ============================ */
