/*
 * FILE:
 * report.c
 *
 * FUNCTION:
 * Fill in widget values based on results of database query
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */

#include <glib.h>
#include <gtk/gtk.h>

#include "perr.h"
#include "util.h"

#include "action.h"
#include "dui-initdb.h"
#include "report.h"
#include "window.h"

struct DuiReport_s
{
	DuiWindow *window;       /* parent window */

   char * row_widgetname;
	GtkWidget *row_iterator; /* get next row */

	GList *columns;          /* list of columns */

	void (*pre) (DuiReport *);
	void (*iter) (DuiReport *);
	void (*set) (DuiReport *, int, int, const char *);
	void (*post) (DuiReport *);
	char **blank_row;

	DuiAction *last_action;	 /* action tht last drew into this window */
};

struct DuiColumn_s
{
	char * col_widgetname;
	GtkWidget *col_widget;
	int col_num;         /* widget - column in which to display the result */
	char *db_field;
	char *key;
	char *value;
};

typedef struct DuiColumn_s DuiColumn;

/* ============================================================ */

DuiReport * 
dui_report_new (DuiWindow *win)
{
	DuiReport * rpt;

	if (!win) return NULL;

	rpt = g_new (DuiReport, 1);

	rpt->window = win;
	rpt->row_widgetname = NULL;
	rpt->row_iterator = NULL;
	rpt->columns = NULL;

	rpt->pre = NULL;
	rpt->iter = NULL;
	rpt->post = NULL;
	rpt->set = NULL;
	rpt->blank_row = NULL;
	
	dui_window_add_report (win, rpt);
	return rpt;
}

void 
dui_report_destroy (DuiReport *rpt)
{
	GList *node;
	if (NULL == rpt) return;

	rpt->window = NULL;
	g_free (rpt->row_widgetname);
	rpt->row_widgetname = NULL;

	rpt->row_iterator = NULL;

	rpt->pre = NULL;
	rpt->iter = NULL;
	rpt->post = NULL;
	rpt->set = NULL;
	g_free (rpt->blank_row);
	rpt->blank_row = NULL;

	for (node=rpt->columns; node; node=node->next)
	{
		DuiColumn *col = node->data;
		if (col->col_widgetname) g_free (col->col_widgetname);
		if (col->db_field) g_free (col->db_field);
		if (col->key) g_free (col->key);
		if (col->value) g_free (col->value);
		g_free (col);
	}
	g_list_free (rpt->columns);
	rpt->columns = NULL;	

	g_free (rpt);
}

/* ============================================================ */

void
dui_report_set_last_action (DuiReport *rpt, DuiAction *act)
{
	if (!rpt) return;
	rpt->last_action = act;
}

/* ============================================================ */

void
dui_report_set_row_widget (DuiReport *rpt, const char *widname)
{
	if (!rpt) return;
	
	if (rpt->row_widgetname) g_free (rpt->row_widgetname);
	if (!widname)
	{
		rpt->row_widgetname = NULL;
	}
	else
	{
		rpt->row_widgetname = g_strdup (widname);
	}
}

/* ============================================================ */

void
dui_report_add_column (DuiReport *rpt, const char * widname, int num, 
                        const char * field, const char * key, 
                        const char * value)
{
	DuiColumn *col;

	if (!rpt) return;
	col = g_new (DuiColumn, 1);

	col->col_widgetname = g_strdup (widname);
	col->col_widget = NULL;
	col->col_num = num;
	col->db_field = g_strdup (field);
	col->key = g_strdup (key);
	col->value = g_strdup (value);
	rpt->columns = g_list_append (rpt->columns, col);
}

/* ============================================================ */
/* ============================================================ */

static void
clist_pre (DuiReport *rpt)
{
	int i, columns;
	GtkCList *clist = GTK_CLIST (rpt->row_iterator);
	
	gtk_clist_freeze (clist);
	gtk_clist_clear (clist);

	/* create a blank row -- hack for adding rows */
	columns = clist->columns;
	if (rpt->blank_row) g_free (rpt->blank_row);
	rpt->blank_row = g_new (char *, columns);
	for (i=0; i<columns; i++) rpt->blank_row[i] = "";
}
	
static void
clist_post (DuiReport *rpt)
{
	GtkCList *clist = GTK_CLIST (rpt->row_iterator);
	
	gtk_clist_thaw(clist);
	g_free (rpt->blank_row);
	rpt->blank_row = NULL;
}
	
static void
clist_iter (DuiReport *rpt)
{
	GtkCList *clist = GTK_CLIST (rpt->row_iterator);
	gtk_clist_append (clist, rpt->blank_row);
}
	
static void
clist_set (DuiReport *rpt, int row, int col, const char * text)
{
	GtkCList *clist = GTK_CLIST (rpt->row_iterator);
	gtk_clist_set_text (clist, row, col, text);
}
	
static void 
clist_sort_column (GtkCList *clist, gint column, DuiReport *rpt)
{
	static GtkSortType direction = GTK_SORT_DESCENDING;

	if (GTK_SORT_DESCENDING == direction) direction = GTK_SORT_ASCENDING;
	else direction = GTK_SORT_DESCENDING;

	gtk_clist_freeze (clist);
	gtk_clist_set_sort_type (clist, direction);
	gtk_clist_set_sort_column (clist, column);
	gtk_clist_sort (clist);

	gtk_clist_thaw (clist);

}

/* ============================================================ */

void
dui_report_show_data (DuiReport *rpt, DuiDBRecordSet *recs)
{
	GList *node;
	int row = 0;
	int more_rows = 1;
	
	if (!rpt) return;

	ENTER ("(rpt=%p, recs=%p)", rpt, recs);
	dui_window_realize (rpt->window);

	/* currently, the only supported widget is clist */
	if (GTK_IS_CLIST(rpt->row_iterator))
	{
		rpt->pre = clist_pre;
		rpt->post = clist_post;
		rpt->iter = clist_iter;
		rpt->set = clist_set;

		gtk_signal_connect (GTK_OBJECT(rpt->row_iterator), 
			"click_column", GTK_SIGNAL_FUNC(clist_sort_column), rpt);
	}
	else
	{
		rpt->pre = NULL;
		rpt->post = NULL;
		rpt->iter = NULL;
		rpt->set = NULL;
	}
	
	if (rpt->pre) (rpt->pre) (rpt);
				
	row = 0;
	
	if (recs) more_rows = dui_recordset_fetch_row (recs);
	while (more_rows)
	{
		if (rpt->iter) (rpt->iter) (rpt);
		for (node=rpt->columns; node; node=node->next)
		{
			DuiColumn *col = node->data;
			const char * val;


			if (recs && col->db_field) 
			{
				val = dui_recordset_get_value (recs, col->db_field);
			}
			else
			if (col->key)
			{
					  DuiInterface *dui = dui_window_get_interface (rpt->window);
					  val = dui_interface_kvp_lookup (dui, col->key);
			}
			else
			val = col->value;

			if (!val) val = "";
			
			if (!col->col_widget || col->col_widget == rpt->row_iterator)
			{
				if (rpt->set) (rpt->set) (rpt, row, col->col_num, val);
			}
			else
			if (GTK_IS_ENTRY (col->col_widget))
			{
				gtk_entry_set_text (GTK_ENTRY(col->col_widget), val);
			}
			else
			if (GTK_IS_LABEL (col->col_widget))
			{
				gtk_label_set_text (GTK_LABEL(col->col_widget), val);
			}
			else
			if (GTK_IS_TEXT (col->col_widget))
			{
				xxxgtk_text_set_text (GTK_TEXT(col->col_widget), val);
			}
			else
			{
				PERR ("unsupported column widget \"%s\"\n", 
				     gtk_type_name(GTK_OBJECT_TYPE(col->col_widget)));
			}
		}

		row ++;
		more_rows = 0;
		if (recs) more_rows = dui_recordset_fetch_row (recs);
	}

	if (rpt->post) (rpt->post) (rpt);

	LEAVE ("(rpt=%p, recs=%p)", rpt, recs);
}

/* ============================================================ */

const char * 
dui_report_get_name (DuiReport *rpt)
{
	if (!rpt) return NULL;
	return dui_window_get_name (rpt->window);
}

/* ============================================================ */

void
dui_report_do_realize (DuiReport *rpt)
{
	GList *node;

	if (!rpt) return;

	ENTER ("(rpt=%p)", rpt);
	rpt->row_iterator = dui_window_get_widget (rpt->window, rpt->row_widgetname);
	
	for (node=rpt->columns; node; node=node->next)
	{
		DuiColumn *col = node->data;

		if (col->col_widgetname)
		{
			col->col_widget = dui_window_get_widget (rpt->window, col->col_widgetname);
		}
	}
	LEAVE ("(rpt=%p)", rpt);
}

/* ============================================================ */

void
dui_report_refresh (DuiReport *rpt)
{
	if (!rpt) return;

	/* no refresh needed if the main window widget doesn't exist */
	if (0 == dui_window_is_realized (rpt->window)) return;

	dui_action_rerun_last_query (rpt->last_action);
}

/* ============================== END OF FILE =================== */
